/* 
 *  Copyright (C) 2001, 2000, 1999,  Rildo Pragana. 
 *  Copyright (C) 1993, 1991  Rildo Pragana.
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *  
 *  You should have received a copy of the GNU General Public License
 *  along with this software; see the file COPYING.  If not, write to
 *  the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 *  Boston, MA 02111-1307 USA
*/ 

/*
**  COBOL scanner 
**  Scanner para o Compilador COBOL
*/

%{
#include "htcoblib.h"
#include "htglobals.h"
#include <ctype.h>

#define MAX_COPY_DEPTH          10
#define MAX_CPYNAME_LEN         65
#define MAXLINEBUF              256

void copy_switch( char *filename );
void scdebug( char *fmt, ... );
#define lexchar input

// extern int lexgetc(void);

int source_debug=0;
int envdiv_bypass_save_sw=0;

static struct copy_symbols {
        struct copy_symbols *next;
        char *value;
        char *replacement;
} *copy_symbols=NULL;
struct copy_symbols *append_copy_symbols=NULL;
int stabs_started=0;

/* COPY file stack data */ 
struct copyfile_info {
  YY_BUFFER_STATE buffer;
  int line;
  struct copy_symbols *last_symbol;
  struct copyfile_info *prev;
};

/* Top of the COPY file stack */
static struct copyfile_info *copyfile_stack = NULL;

extern FILE *o_lst;
int lex_fgets( char *buf, int maxsize );

#define YY_INPUT(buf,result,max_size) \
        { \
                result = lex_fgets( buf, max_size ); \
        }

// redefines the lexer so that we define yylex to call yylex2 and save
// the token returned by the real lexer (yylex2)
int prev_token=0;
#define YY_DECL int yylex2 YY_PROTO((void))
%}

letters [A-Za-z]
alphanum [A-Za-z0-9]
ident {alphanum}(({alphanum}|-|_|$)*{alphanum}+)?
exletter {letters}|-
digit [0-9]
pic_ch [9XVSBZAP.,-+*/$]
relop [<>=]
white [\r\t ]
white_nl [\r\t \n]
lpar \(
rpar \)
sdelim \"
ldelim \'
lhexa [xX]
any [!-~]
int {digit}({digit})*
decdelim [.,]
sign [+-]
notperiod [^.\n]
period [.]

%{
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include "htcoboly.h"
#include "htcobol.tab.h"    /* definitions from YACC */
#include "reswords.h"

#define TOKTEXT_SIZE 1024

char toktext[TOKTEXT_SIZE];
char *tp;
unsigned char hnum[512];
unsigned int varnumber;
int reserved_token;
int previous_state = 0;
short function_seen = 0;
int eos_detect=0;

#define svtoken() strncpy(toktext,yytext,sizeof(toktext))  

%}

%x IDENT_ST ENVIR_ST DATA_ST COMMENT_ST FD_ST REDEF_ST
%x SUBSCRIPTS_ST COPYFILE_ST EXCEPTION_ST
%x PIC_ST PIC_ST2 COPY_ST COMMENT1_ST

%%
%{
        extern int curr_division;
        extern int need_subscripts;
        struct copy_symbols *tmp = NULL;

        switch (curr_division) {
        case CDIV_IDENT:
                scdebug("-> IDENT_ST\n");
                BEGIN IDENT_ST;
                break;
        case CDIV_ENVIR:
                scdebug("-> ENVIR_ST\n");
                BEGIN ENVIR_ST;
                break;
        case CDIV_DATA:
                scdebug("-> DATA_ST\n");
                BEGIN DATA_ST;
                break;
        case CDIV_PROC:
                //stabs_started=1;
                scdebug("-> INITIAL\n");
                BEGIN INITIAL;
                break;
        case CDIV_COMMENT:
                scdebug("-> COMMENT_ST\n");
                BEGIN COMMENT_ST;
                break;
        case CDIV_COMMENT1:
                scdebug("-> COMMENT1_ST\n");
                BEGIN COMMENT1_ST;
                break;
        case CDIV_FD:
                scdebug("-> FD_ST\n");
                BEGIN FD_ST;
                break;
        case CDIV_REDEF:
                scdebug("-> REDEF_ST\n");
                BEGIN REDEF_ST;
                break;
        case CDIV_SUBSCRIPTS:
                scdebug("-> SUBSCRIPTS_ST\n");
                BEGIN SUBSCRIPTS_ST;
                break;
        case CDIV_COPYFILE:
                append_copy_symbols=NULL;
                scdebug("-> COPYFILE_ST\n");
                previous_state = YY_START;
                BEGIN COPYFILE_ST;
                break;
        case CDIV_EXCEPTION:
                scdebug("-> EXCEPTION_ST\n");
                BEGIN EXCEPTION_ST;
                break;
        case CDIV_PIC:
                scdebug("-> PIC_ST\n");
                BEGIN PIC_ST;
                break;
        case CINITIAL:
                scdebug("-> INITIAL\n");
                BEGIN INITIAL;
                break;
        }
        curr_division=0; /* avoid new state switch */
%}

<*>^"#".* {
  char *endp;
  source_lineno = strtol (yytext + 2, &endp, 10);
  if (source_lineno > 0)
    source_lineno--;
  if (source_filename)
    free (source_filename);
  source_filename = strdup (strchr (endp, '"') + 1);
  strrchr (source_filename, '"')[0] = '\0';
  scdebug("NEW FILE/LINE : %s,%d\n", source_filename,source_lineno);
}

<*>"EJECT"|"SKIP1"|"SKIP2"|"SKIP3" { /* throw away this */ }

<*>^[ \t]?"*".*$        { 
           /* allow floating comment line */  
                scdebug("<comment>: %s:\n", yytext);
        }

<*>^"D ".*      { 
                if (!source_debug)
                        scdebug("<comment>: %s:\n", yytext);
                else
                        yyless(2);
        }

";" { }
"," { }

{lpar} {
                if (need_subscripts) {
                        scdebug("<LPAR>:\n");
                        need_subscripts=0;
                        return(LPAR);
                }
                else {
                        return('('); 
                }
        }
{white}+ { }

<INITIAL,SUBSCRIPTS_ST>{ident} { {
                struct sym *symbol;
                svtoken();
                if (is_integer_token(toktext) && prev_token == PERIOD_TOK) {
                        yylval.sval=install(toktext,SYTB_LAB,0); 
                        scdebug("<LABELSTR>: %s\n", toktext);
                        return(LABELSTR);
                } 
                /* if (is_integer_token(toktext) && !eos_detect) {*/
                if (is_integer_token(toktext)) {
                        scdebug("<NLITERAL>: %s\n", toktext);
                        yylval.lval = (struct lit *)install_lit(toktext,0,0);
                        return(NLITERAL);
                }
                eos_detect=0;
                if (!is_valid_identifier(toktext)) {
                        yylval.sval=install(toktext,SYTB_LAB,0); 
                        scdebug("<LABELSTR>: %s\n", toktext);
                        return(LABELSTR);
                } 
                else {
                        if ((reserved_token=reserved(toktext))!=0) 
                        {
                                if (reserved_token == FUNCTION) function_seen=TRUE;
                                /* special case of 4 intrinsics which are also reserved */
                                if (function_seen && lookup_intrinsic_sym(toktext) != NULL)
                                {
                                        yylval.sval=install(toktext,SYTB_LAB,0); 
                                        scdebug(" <LABELSTR>: %s\n", toktext);
                                        return(LABELSTR);
                                }
                                scdebug(" <reserved>: %s\n", toktext);
                                return reserved_token;
                        }
                        /* if symbol was defined at the data division */
                        if (symbol=lookup(toktext,2)) {
                                yylval.sval=symbol;
                                if (symbol->type == '8') {
                                        scdebug("<VARCOND>: %s\n",toktext);
                                        update_xreflist(yylval.sval); 
                                        return(VARCOND);
                                }
                                else {
                                        if (is_subscripted(symbol)) {
                                                        scdebug("<SUBSCVAR>: %s\n",toktext);
                                                        return(SUBSCVAR);
                                        }
                                        scdebug("<VARIABLE>: %s\n", toktext);
                                        update_xreflist(yylval.sval); 
                                        return(VARIABLE);
                                }
                        } 
                        /* otherwise, it should be a label */
                        else {
                                yylval.sval=install(toktext,SYTB_LAB,0); 
                                scdebug("<LABELSTR>: %s\n", toktext);
                                return(LABELSTR);
                        }
                }
        } }

<ENVIR_ST>{ident} {
                svtoken();
                if (is_integer_token(toktext)) {
                        scdebug("<NLITERAL>: %s\n", toktext);
                        yylval.lval = (struct lit *)install_lit(toktext,0,0);
                        return(NLITERAL);
                }
                if (!is_valid_identifier(toktext)) {
                        yylval.sval=install(toktext,SYTB_LAB,0); 
                        scdebug("<LABELSTR>: %s\n", toktext);
                        return(LABELSTR);
                }
                else {
                      if ((reserved_token=reserved(toktext))!=0) 
                      {
                         scdebug(" <reserved>: %s\n", toktext);
                         return reserved_token;
                      }
                      scdebug("<STRING>: %s\n", toktext);
                      if (envdiv_bypass_save_sw == 0) {
                         yylval.sval=install(toktext,SYTB_VAR,0);
                         update_xreflist(yylval.sval); 
                      }
                      return(STRING);
                }

}

<EXCEPTION_ST>{ident} { {
                struct sym *symbol;
                svtoken();
                if (is_integer_token(toktext)) {
                        scdebug("<NLITERAL>: %s\n", toktext);
                        yylval.lval = (struct lit *)install_lit(toktext,0,0);
                        return(NLITERAL);
                }
                if (!is_valid_identifier(toktext)) {
                        yylval.sval=install(toktext,SYTB_LAB,0); 
                        scdebug("<LABELSTR>: %s\n", toktext);
                        return(LABELSTR);
                }
                else {
                        if ((reserved_token=reserved(toktext))!=0) 
                        {
                                scdebug(" <reserved>: %s\n", toktext);
                                if (reserved_token == NOT)
                                        reserved_token = NOTEXCEP;
                                return reserved_token;
                        }
                        /* if symbol was defined at the data division */
                        if (symbol=lookup(toktext,2)) {
                                yylval.sval=symbol;
                                if (symbol->type == '8') {
                                        scdebug("<VARCOND>: %s\n",toktext);
                                        update_xreflist(yylval.sval); 
                                        return(VARCOND);
                                }
                                else {
                                        scdebug("<VARIABLE>: %s\n", toktext);
                                        update_xreflist(yylval.sval); 
                                        return(VARIABLE);
                                }

                        } 
                        /* otherwise, it should be a label */
                        else {
                                yylval.sval=install(toktext,SYTB_LAB,0);
                                scdebug("<LABELSTR>: %s\n", toktext);
                                return(LABELSTR);
                        }
                }
        } }

<FD_ST>{
{ident} {
                svtoken();
                if (is_integer_token(toktext)) {
                        scdebug("<NLITERAL>: %s\n", toktext);
                        yylval.lval = (struct lit *)install_lit(toktext,0,0);
                        return(NLITERAL);
                }
                if (!is_valid_identifier(toktext)) {
                        yylval.sval=install(toktext,SYTB_LAB,0); 
                        scdebug("<LABELSTR>: %s\n", toktext);
                        return(LABELSTR);
                }
                else {
                        if ((reserved_token=reserved(toktext))!=0) 
                        {
                                scdebug(" <reserved>: %s\n", toktext);
                                return reserved_token;
                        }
                        scdebug("<STRING>: %s\n", toktext);
                        /* install w/clone if neccessary */
                        if ((yylval.sval=lookup(toktext,SYTB_VAR))==NULL)
                                yyerror("undefined %s at FD/SD",toktext); 
                        return(STRING);
                }
        }
[ \t]* { }
}

<REDEF_ST>{ident} {
                svtoken();
                if (is_integer_token(toktext)) {
                        scdebug("<NLITERAL>: %s\n", toktext);
                        yylval.lval = (struct lit *)install_lit(toktext,0,0);
                        return(NLITERAL);
                }
                if (!is_valid_identifier(toktext)) {
                        yylval.sval=install(toktext,SYTB_LAB,0); 
                        scdebug("<LABELSTR>: %s\n", toktext);
                        return(LABELSTR);
                }
                else {
                        if ((reserved_token=reserved(toktext))!=0) 
                        {
                                scdebug(" <reserved>: %s\n", toktext);
                                return reserved_token;
                        }
                        scdebug("<STRING>: %s\n", toktext);
                        /* don't clone redefined var */
                        yylval.sval=install(toktext,SYTB_VAR,0); 
                        update_xreflist(yylval.sval); 
                        return(STRING);
                }
        }

<DATA_ST>{ident} {
                svtoken();
                if (is_integer_token(toktext)) {
                        scdebug("<NLITERAL>: %s\n", toktext);
                        yylval.lval = (struct lit *)install_lit(toktext,0,0);
                        return(NLITERAL);
                }
                if (!is_valid_identifier(toktext)) {
                        yylval.sval=install(toktext,SYTB_LAB,0); 
                        scdebug("<LABELSTR>: %s\n", toktext);
                        return(LABELSTR);
                }
                else {
                        if ((reserved_token=reserved(toktext))!=0) 
                        {
                                scdebug(" <reserved>: %s\n", toktext);
                                return reserved_token;
                        }
                        scdebug("<STRING>: %s\n", toktext);
                        /* install w/clone if neccessary */
                        yylval.sval=install(toktext,SYTB_VAR,1);
                        update_xreflist(yylval.sval); 
                        return(STRING);
                }
        }

<SUBSCRIPTS_ST>{
";" { }
"+"{int} {
                svtoken();
                scdebug("<NLITERAL>: %s\n", toktext);
                yylval.lval = (struct lit *)install_lit(&toktext[1],0,0);
                return(NLITERAL);
	} 

{any}   {
                svtoken();
                scdebug("<0: any>: %s\n", toktext);
                return(toktext[0]); 
        }
{white}+ { }
}       

<IDENT_ST>{
[ \t]* { }
\./[ \t\n]+ { 
                scdebug("<PERIOD_TOK>:\n");
                return(PERIOD_TOK); 
        }
{ident} {
                svtoken();
                if (is_integer_token(toktext)) {
                        scdebug("<NLITERAL>: %s\n", toktext);
                        yylval.lval = (struct lit *)install_lit(toktext,0,0);
                        return(NLITERAL);
                }
                if (!is_valid_identifier(toktext)) {
                        yylval.sval=install(toktext,SYTB_LAB,0); 
                        scdebug("<LABELSTR>: %s\n", toktext);
                        return(LABELSTR);
                }
                else {
                        if ((reserved_token=reserved(toktext))!=0) 
                        {
                                scdebug(" <reserved>: %s\n", toktext);
                                return reserved_token;
                        }
                        scdebug(" <IDSTRING>: %s\n", toktext);
                        yylval.str=toktext;
                        return IDSTRING;
                }
        }
{any}   {
                svtoken();
                scdebug("<0: any>: %s\n", toktext);
                return(toktext[0]); 
        }
"\n" { }
"," { }
";" { }
} /* end of <IDENT_ST> */

<COPY_ST>{
\./[ \t]*"\n" { 
                scdebug("<PERIOD_TOK>:\n");
                copy_switch(include_filename);
                scdebug("<-> previous_state>\n");
                BEGIN(previous_state);
                return(PERIOD_TOK); 
        }
[ \t]*","[ \t]* { scdebug("<COPY_ST: sep>: %s\n", toktext);  }
"==" {
                int eq_seen=0;
                tp=toktext;
                while (1) {
                        while((*tp=lexchar())!='=') { tp++; eq_seen=1; }
                        if ((*tp=lexchar())=='=' && eq_seen) {
                                *tp=0;
                                break;
                        }
                }
                scdebug(" <IDSTRING>: %s\n", toktext);
                yylval.str = savename(toktext);
                return IDSTRING;
        }
{ident} {
                if ((reserved_token=reserved(yytext))!=0) 
                {
                   if (reserved_token==BY) {
                                scdebug(" <reserved>: %s\n", yytext);
                                return BY; 
                   }
                   if (reserved_token==REPLACING) {
                                scdebug(" <reserved>: %s\n", yytext);
                                return REPLACING; 
                   }
                }
                scdebug(" <IDSTRING>: %s\n", yytext);
                yylval.str = savename(yytext);
                return IDSTRING;
        }
{white}+ { }

} /* end of COPY_ST */

<COPYFILE_ST>{
{any}+{alphanum} {
                scdebug(" <IDSTRING>: %s\n", yytext);
                yylval.str = savename(yytext);
                scdebug("-> COPY_ST\n"); 
                BEGIN(COPY_ST);
                return IDSTRING;
  }
{white}+ { }
}

<INITIAL,ENVIR_ST,DATA_ST,REDEF_ST,EXCEPTION_ST>{

";" { }
\./[ \t\n]+ { 
                scdebug("<PERIOD_TOK>:\n");
                eos_detect++;
                return(PERIOD_TOK); 
        }

{sdelim}        {
                tp=toktext;
                while(1) {
                        while((*tp=lexchar())!='"') {
                                if (tp >= toktext+TOKTEXT_SIZE) {
                                        yyerror("scanner error: token too large");
                                        while(lexchar()!='"') ;
                                        *tp=0;
                                        scdebug("<CLITERAL>: %s\n", toktext);
                                        yylval.lval = (struct lit *)install_lit(toktext,0,0);
                                        return(CLITERAL);
                                }
                                tp++;
                        }
                        /* replace two consecutives '"' by one and continue */
                        if ((*tp=lexchar())=='"') {
                                tp++;
                                continue;
                        }
                        else {
                                unput(*tp);
                                break;
                        }
                }
                *tp=0;
                scdebug("<CLITERAL>: %s\n", toktext);
                yylval.lval = (struct lit *)install_lit(toktext,0,0);
                return(CLITERAL);
        }

{lhexa}{ldelim}       { {
                unsigned char *s=hnum;
                int len;
                tp=toktext;
                while((*tp=toupper(lexchar()))!='\'') {
                                if (tp >= toktext+TOKTEXT_SIZE) {
                                        yyerror("scanner error: token too large");
                                        while(lexchar()!='\'') ;
                                        *tp=0;
                                        scdebug("<CLITERAL>: %s\n", toktext);
                                        yylval.lval = (struct lit *)install_lit(toktext,0,0);
                                        return(CLITERAL);
                                }
                        *s = (*tp > '9') ?
                                (*tp - 'A' + 0x0a) : (*tp - '0');
                        if ((*tp=toupper(lexchar()))=='\'') {
                                break;
                        }
                        *s++ = *s * 0x10 + ((*tp > '9') ? 
                                (*tp - 'A' + 0x0a) : (*tp - '0'));
                }
                len = (int) (s - hnum);
                yylval.lval = (struct lit *)install_lit(hnum,len,0);
                scdebug("<CLITERAL>: %s\n", toktext);
                return(CLITERAL);
        } }

{lhexa}{sdelim}       { {
                unsigned char *s=hnum;
                int len;
                tp=toktext;
                while((*tp=toupper(lexchar()))!='\"') {
                                if (tp >= toktext+TOKTEXT_SIZE) {
                                        yyerror("scanner error: token too large");
                                        while(lexchar()!='\"') ;
                                        *tp=0;
                                        scdebug("<CLITERAL>: %s\n", toktext);
                                        yylval.lval = (struct lit *)install_lit(toktext,0,0);
                                        return(CLITERAL);
                                }
                        *s = (*tp > '9') ?
                                (*tp - 'A' + 0x0a) : (*tp - '0');
                        if ((*tp=toupper(lexchar()))=='\"') {
                                break;
                        }
                        *s++ = *s * 0x10 + ((*tp > '9') ?
                                (*tp - 'A' + 0x0a) : (*tp - '0'));
                }
                len = (int) (s - hnum);
                yylval.lval = (struct lit *)install_lit(hnum,len,0);
                scdebug("<CLITERAL>: %s\n", toktext);
                return(CLITERAL);
        } }

{ldelim}        {
                tp=toktext;
                while(1) {
                        while((*tp=lexchar())!='\'') { 
                                if (tp >= toktext+TOKTEXT_SIZE) {
                                        yyerror("scanner error: token too large");
                                        while(lexchar()!='\'') ;
                                        *tp=0;
                                        scdebug("<CLITERAL>: %s\n", toktext);
                                        yylval.lval = (struct lit *)install_lit(toktext,0,0);
                                        return(CLITERAL);
                                }
                                tp++;
                        }
                        /* replace two consecutives '\'' by one and continue */
                        if ((*tp=lexchar())=='\'') {
                                tp++;
                                continue;
                        }
                        else {
                                unput(*tp);
                                break;
                        }
                }
                *tp=0;
                yylval.lval = (struct lit *)install_lit(toktext,0,0);
                scdebug("<CLITERAL>: %s\n", toktext);
                return(CLITERAL);
        }

{digit}*{decdelim}?{int}+       {
                svtoken();
                scdebug("<NLITERAL>: %s\n", toktext);
                yylval.lval = (struct lit *)install_lit(toktext,0,0);
                return(NLITERAL);
        }
%{ /*
{digit}({exletter}|{digit})* { 
        svtoken();
        scdebug("<LABELSTR>: %s\n", toktext);
        yylval.sval=install(toktext,SYTB_LAB,0);
        return(LABELSTR); 
        }
*/ 
%}

{white}+ { }


{relop}{relop}? {
                svtoken();
                if ((reserved_token=reserved(toktext))!=0) {
                        scdebug(" <reserved>: %s\n", toktext);
                        return reserved_token;
                }
                yyerror("relational operator unknown");
        }

{lpar}  {
                svtoken();
                if (function_seen) {
                        function_seen=FALSE;
                        }
                scdebug("<0: lpar>: %s\n", toktext);
                return(toktext[0]); 
        }
"**"    {
                svtoken();
                scdebug("<0: POW_OP>: %s\n", toktext);
                return(POW_OP); 
        }
"," { }
";" { }
{any}   {
                svtoken();
                scdebug("<0: any>: %s\n", toktext);
                return(toktext[0]); 
        }

"\n"   { }      

} /* end of <INITIAL,DATA_ST,ENVIR_ST> */

<COMMENT_ST>{
.*     { 
         svtoken();
         scdebug(" <COMMENT_ST>: %s\n", toktext);
       }
"\n"   { BEGIN INITIAL; }
} /* end of COMMENT_ST */

<COMMENT1_ST>{
{notperiod}* { 
         svtoken();
         scdebug(" <COMMENT1_ST>:>>%s<<:\n", toktext);
        }
"."" "*/"\n"   { 
         svtoken();
         scdebug(" <COMMENT1_ST>:>>%s<<:\n", toktext);
         BEGIN INITIAL; 
       }
"."    { 
         svtoken();
         scdebug(" <COMMENT1_ST>:>>%s<<:\n", toktext);
       }
"\n"   { 
         svtoken();
         scdebug(" <COMMENT1_ST>:<NL>:\n");
       }
} /* end of COMMENT1_ST */

<PIC_ST>{white_nl}* { 
                BEGIN PIC_ST2;
                scdebug("<PICS -> PIC_ST2>:\n");
        }

<PIC_ST2>{
{white_nl}* { 
                BEGIN DATA_ST;
                scdebug("<PICS -> DATA_ST>:\n");
        }
\./[ \t\n]+ { 
                BEGIN DATA_ST;
                scdebug("<PICS -> DATA_ST>: PERIOD_TOK\n");
                return PERIOD_TOK;
        }
}

<PIC_ST,PIC_ST2>{
{lpar}{digit}({digit})*{rpar} { 
                { 
                char *s;
                yylval.ival = 0;
                s = yytext+1; /* bypass lpar */
                while (*s!=')')
                        yylval.ival = 10 * yylval.ival + *s++ - '0';
                scdebug("<PICS: MULTIPLIER>: %d\n", yylval.ival);
                return(MULTIPLIER); 
                } 
        }
{ident} {
                svtoken();
                if ((reserved_token=reserved(toktext))!=0) 
                {
                        scdebug(" <reserved>: %s\n", toktext);
                        BEGIN PIC_ST;
                        scdebug("<PIC_ST2-> PIC_ST>:\n");
                        return reserved_token;
                }
                REJECT;
        }
";" { }
{any}   { 
                scdebug("<PICS: any char>: %c\n", yytext[0]);
                yylval.ival=(int)yytext[0];
                return(CHAR); }
"\n" { }
} /* end of PICS */

<<EOF>> {
        int tmp;
        struct copy_symbols *cs,*cs1;
        if (!copyfile_stack) {
                //fprintf(stderr,"scanner receiving <<EOF>>\n");
                yyterminate();
        }
        else {
                if (HTG_verbose == TRUE )
                    printf("Lines included in file: %4d\n", source_lineno);

                tmp = source_lineno;
                yy_delete_buffer( YY_CURRENT_BUFFER );
                yy_switch_to_buffer( copyfile_stack->buffer );
                source_lineno = copyfile_stack->line;
                /* free all replacements in this dead context */
                cs1 = copyfile_stack->last_symbol;
                cs = copy_symbols;
                if (cs == cs1)
                        copy_symbols = NULL;
                while (cs) {
                        if (cs->next == cs1)
                                break;
                        cs = cs->next;
                }
                if (cs)
                        cs->next = NULL; 
                if (cs1) {
                        while (cs1) {
                                if (cs1-> value ) free( cs1->value );
                                if (cs1->replacement ) free( cs1->replacement );
                                cs = cs1->next;
                                free( cs1 );
                                cs1 = cs;
                        }
                }
                source_lineno = source_lineno + tmp;

		/* Free the top copy file info */
		{
		  struct copyfile_info *top = copyfile_stack;
		  copyfile_stack = copyfile_stack->prev;
		  if ( top ) free (top);
		}
        }
}

.	{svtoken();
         scdebug("invalid character:>>%x<<:\n", (unsigned char)toktext[0]);
         yyerror("invalid character:>>%x<<:\n", (unsigned char)toktext[0]);
	}
%%

int yylex YY_PROTO((void))
{
	prev_token = yylex2();
	/*printf("tok:%x\n", prev_token);*/
	return prev_token;
}

int is_integer_token( char *tok ) {
        while (isdigit(*tok)) tok++;
        if (*tok=='\0') return 1;
        return 0;
}

int is_valid_identifier( char *tok ) {
        while (*tok && !isalpha(*tok)) tok++;
        if (*tok=='\0') {
                return 0;
        }
        return 1;
}

int lex_fgets( char *buf, int maxsize ) {
        int result;
        char *s;
        s = fgets( buf,maxsize,yyin );
        result = (s==NULL) ? YY_NULL: strlen(s);
//	result = fread( buf, 1, maxsize, yyin);
        if (result) { 
//              fprintf(stderr,"lexbuf before: |%s|\n",buf); 
//              fprintf(stderr,"lexbuf before: |%s|",buf); 
//XXX                copy_do_replacements( buf ); 
                result = strlen(buf);
//              fprintf(stderr,"lexbuf after: |%s|\n",buf); 
//              fprintf(stderr,"lexbuf after: |%s|",buf); 
//              if (HTG_list_flag) 
//                      fprintf( o_lst,"%04d:%s",++source_lineno,buf ); 
//              else 
//                      ++source_lineno; 
                ++source_lineno; 
        } 
        return result;
}

void dcs() {
  fprintf(stderr,"===============\n");
  if (copyfile_stack)
    fprintf(stderr,"last symbol: %s\n",
	    copyfile_stack->last_symbol ? 
	    copyfile_stack->last_symbol->value : "");
  if (copy_symbols) {
    struct copy_symbols *ccs;
        ccs= copy_symbols;
        while (ccs) {
	  fprintf(stderr,"copy symbolA: %s\n",ccs->value);
	  ccs = ccs->next;
        }
  }
}

/* despite the name, copy_strip_spaces eliminates ", " and "; "
        separators too, besides multiples spaces.
        This follows standard CD-1.2 pp.679, rule 7.b */
char *
copy_strip_spaces( char *s ) {
        char *s1 = NULL,*new = NULL;
        s1 = new = malloc(strlen(s)+2);
        while (*s==' ' || *s=='\t') /* don't strip indenting */
                *s1++=*s++;
        while (*s) {
                if ((*s==',' || *s==';') && *(s+1)==' ') s++;
                if (*s==' ' || *s=='\t' || *s=='\n') {
                        *s1++ = ' ';
                        if (*s=='\n') break;
                        while (*s && (*s==' ' || *s=='\t')) s++;
                }
                else
                        *s1++ = *s++;
        }
        *s1=0;
        return new;
}

/*void add_copy_replacement( char *orig_value, char *replacement ) {
        struct copy_symbols *cs,*tmp;
        char *value;
//fprintf(stderr,"add_copy_replacement: \"%s\" BY \"%s\"\n",
//              orig_value,replacement);
        value=copy_strip_spaces(orig_value);
        free(orig_value);
        cs = malloc(sizeof(struct copy_symbols));
        cs->value = value;
        cs->replacement = replacement;
        cs->next = NULL;
        if (append_copy_symbols==NULL) 
                append_copy_symbols = cs;
        else {
                tmp=append_copy_symbols;
                while (tmp->next) 
                        tmp=tmp->next;
                tmp->next = cs;
        }
}*/

/*void copy_do_replacements( char *buffer ) {
        struct copy_symbols *cs;
        char *stripped,*s,*tmp,*s1,c;
        
        // Replacement table empty
        if (copy_symbols==NULL) return;

        // skip comments
        if (*buffer == '*')  {
//         fprintf(stderr,"compare ignore comments:%s:\n", buffer);
           return;
        }

        s1 = tmp = malloc(MAXLINEBUF);
        s = stripped = copy_strip_spaces( buffer );
        while (*s && (*s == ' ' || *s == '\t')) /* bypass white *//* 
                *s1++ = *s++; 
        while (*s) {
                for (cs=copy_symbols;cs;cs=cs->next) {
                        if (!strncmp(s,cs->value,strlen(cs->value))) {
                                c = *(s+strlen(cs->value));
                                if (c!=' ' && c!='\0') continue;
                                strcpy(s1,cs->replacement);
                                s += strlen(cs->value);
                                s1 += strlen(cs->replacement);
                                *s1++ = ' ';
                                break;
                        }
                }
                if (cs==NULL) {
                        while (*s && *s != ' ' && *s != '\t') 
                                *s1++ = *s++;
                        while (*s && (*s == ' ' || *s == '\t' || *s == '\n')) {
                                if (*s == '\n') {
                                        *s1++ = *s++;
                                        break;
                                }
                                *s1++ = *s++;
                        }
                }
        }
        *s1++='\n';
        *s1++=0;
        strcpy(buffer,tmp);
        free(stripped);
        free(tmp);
}
*/
void copy_switch( char *filename ) {
        struct copy_symbols *tmp = NULL;
	struct copyfile_info *ci = NULL;
	char fullname[FILENAME_MAX];

	/* Allocate new copy file info */
	ci = malloc (sizeof (struct copyfile_info));
        ci->buffer = YY_CURRENT_BUFFER;
        ci->line = source_lineno;
        if ((tmp=copy_symbols)!=NULL) {
                while (tmp->next) tmp=tmp->next;
                tmp->next = append_copy_symbols;
        }
        else {
                copy_symbols = append_copy_symbols;
        }
        ci->last_symbol = append_copy_symbols;
	ci->prev = copyfile_stack;
	copyfile_stack = ci;

        if (find_copybook_file(filename, fullname) != 0) {
	  fprintf(stderr, "error: copy file '%s' not found\n", filename);
	  fprintf(stderr, "Include(copybooks) search path(s)=%s;\n", HTG_COPYDIR);
	  exit(1);
        }

        if (HTG_verbose == TRUE ) {
            printf("Including copybook: %s\n",fullname);
        }

//      printf("Including: %-20s\n",filename);
        if ((yyin = fopen( fullname, "r" ))==NULL) {
                fprintf(stderr, "Cannot open copy file %s\n", fullname);
                exit(1);
        }

        source_lineno = 0;
        yy_switch_to_buffer( yy_create_buffer(yyin,YY_BUF_SIZE) );
}

//char *savename( char *name );
//int hash( char *name );


#if defined(SunOS)
va_list __builtin_va_alist;
#endif

void scdebug( char *fmt, ... ) {
        va_list args;
        va_start(args,fmt);
#ifdef DEBUG_SCANNER
        fprintf(stderr,"SCAN(%4d):",source_lineno);
        vfprintf(stderr,fmt,args);
#endif
        va_end(args);
}

int yywrap() { return 1; }

/* end of scan.l */
