.\"
.\"  Created by Ferran Pegueroles using help2man and modified by hand.
.\"  Modified and updated by David Essex.
.\" 
.TH HTCOBOL 1 "January 22, 2002"
.UC 6
.SH NAME
htcobol \- COBOL 85 compiler
.SH SYNOPSIS
.B htcobol 
[
.I options
] 
.I filename
.SH "DESCRIPTION"
A compiler for the 
\fBCO\fRmmon
\fBB\fRusiness 
\fBO\fRriented 
\fBL\fRanguage, 
\fBCOBOL\fR.
.PP 
.I Htcobol
reads a COBOL source in the file
.I filename
and depending on the option, will preprocess, compile, assemble and link 
to generate an executable binary.
.PP
The compiler generates GNU assembler for the IA32 (i386) platforms.
With the aid of the \fBGCC\fR 
tool set, this intermediate code can then be assembled and linked to create an executable binary.  
.PP
A executable binary can be created either directly by the compiler, or by 
generating intermediate assembler code and using a 
.I Makefile
for the assemble and link steps. 
.PP
The compiler recognizes several command line options as described below.
.PP
You can get a help message by invoking htcobol with the
.B \-h
option.  
.PP
.SH "INITIALIZATION FILES"
Many compiler options can be set using the 
resource file and/or command line options.
.PP
Default resource file name is
.I htcobolrc\fR.
.PP
The precedence of any compiler option is as follows:
.TP
1.
Command line options, if available.
.TP
2.
Environment variables, if available. 
.TP
3.
The resource file options, if available.
.TP
4.
Compile time resource default options, if available.
.SH "OPTIONS"
.B Compiler specific options:
.TP
\fB\-h\fR 
Display help.
.TP 
\fB\-a\fR
Create static library; Preprocess, compile, assemble and archive
.TP 
\fB\-B\fR
mode Specify binding mode (static/dynamic)
.TP 
\fB\-c\fR
Compile to a statically linked object module
.TP 
\fB\-E\fR
Output preprocessor to standard output only; do not compile, assemble or link
.TP 
\fB\-g\fR
Generate compiler debugging output
.TP 
\fB\-l\fR <name> 
Add library name to link step
.TP 
\fB\-L\fR <dir> 
Add directory to library search path
.TP 
\fB\-m\fR 
Create shared library; Preprocess, compile, assemble and link
.TP 
\fB\-n\fR
Don't actually run any commands; just print them
.TP 
\fB\-o\fR <file> 
Specify output file name
.TP 
\fB\-S\fR
Preprocess, compile (generate assembler code) only; do not assemble or link
.TP 
\fB\-t\fR
Doesn't remove the intermediary files (assembly file, pre-processed COBOL file) generated 
during the compilation.
.TP 
\fB\-v\fR
Generate verbose compiler output
.TP 
\fB\-V\fR
Display compiler version information and exit
.TP 
\fB\-Wl,<options> \fR 
Pass comma-separated <options> on to the linker
.TP 
\fB\-x\fR
Compile to an executable module
.TP 
\fB\-z\fR
Generate very verbose compiler output
.PP 
.B COBOL specific options:
.TP 
\fB\-C\fR
Make all COBOL calls dynamic
.TP 
\fB\-D\fR
Include source debugging lines
.TP 
\fB\-F\fR
Input source is in standard fixed column format
.TP 
\fB\-I\fR <path> 
Define include (copybooks) search path(s) (default \-I./)
The path may be either a single directory, or a list of 
directories separated by a `:' (`;' on the Win32 platform).
.TP 
\fB\-P\fR
Generate output listing file
.TP 
\fB\-T\fR <num>
Expand tabs to number of space(s) (default T=8)
.TP 
\fB\-X\fR
Input source is in X/Open free format (default format)
.PP 
.SH "FILES"
.TP
.I htcobolrc\fR resource options file.
.SH ENVIRONMENT
.TP
.B TCOB_OPTIONS
.TP
Name of resource options file. 
.TP
Default name is 
.I htcobolrc\fR.
.TP
.B TCOB_OPTIONS_PATH
Options file directory path.
.TP
.B TCOB_PP_PATH
Preprocessor directory path.
.TP
.B TCOB_LD_LIBRARY_PATH
Dynamically loaded libraries search path.
.TP
.B  TEMP
Temporary files directory path.
.SH "SEE ALSO"
GCC(1), as(1), ld(1), make(1).
