// 
// Copyright (C) 2001, 2000, 1999,  Rildo Pragana, Jim Noeth, 
//               Andrew Cameron, David Essex.
// Copyright (C) 1993, 1991  Rildo Pragana.
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation; either version 2.1,
// or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; see the file COPYING.LIB.  If
// not, write to the Free Software Foundation, Inc., 59 Temple Place,
// Suite 330, Boston, MA 02111-1307 USA


//
//  TinyCOBOL Run Time Library 
//  Dynamic CALL statement functions
//

//#define USE_LTDL_LIB 1

#define SUBR_HASHLEN                     128
#define HTCOB_PATH_PWD_STR                "."

#include "htcoblib.h"
#include "rterrors.h"

#ifdef __MINGW32__
# ifndef USE_LTDL_LIB
#  define USE_LTDL_LIB 1
# endif
# include "ltdl.h"
#else
# ifdef USE_LTDL_LIB
#  include <ltdl.h>
# endif
# include <dlfcn.h>
# include <glob.h>
#endif // MINGW32


#ifdef __WINDOWS__
# define HTCOB_DYNLIBS_PATTERN           "*.dll"
# define HTCOB_DYNLIBS_SUFIX             ".dll"
# define HTCOB_PATH_SEP_CHAR                ';'
# define HTCOB_PATH_DELIM_STR                "\\"
#else
# define HTCOB_DYNLIBS_PATTERN           "*.so"
# define HTCOB_DYNLIBS_SUFIX             ".so"
# define HTCOB_PATH_SEP_CHAR                ':'
# define HTCOB_PATH_DELIM_STR                "/"
#endif

#ifdef USE_LTDL_LIB
static int dl_initialized = 0;
#endif /* USE_LTDL_LIB  */

static char subrname[65]; /* allow at most 64 chars subr names */

static struct resolve_handle {
        void *handle;
        struct resolve_handle *next;
} *resolve_handles=NULL;

/* 
   this mini-symtable will register previously called subroutines 
   to speed up resolving their names to pointers 
*/
static struct subr_desc {
        struct subr_desc *next;
        char *name;
        void *(*subr)();
} *subrtab[ SUBR_HASHLEN ]={NULL};

#ifndef __MINGW32__
static char libs_pattern[]=HTCOB_DYNLIBS_PATTERN;
#endif

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                          subr_hash                                     |
 |                                                                        |
\*------------------------------------------------------------------------*/

static int subr_hash( char *s ) {
        int val=0, i;
        for (i=0; s[i]!='\0'; i++)
                val += s[i];
        return( val % SUBR_HASHLEN );
}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                          savename                                      |
 |                                                                        |
\*------------------------------------------------------------------------*/

static char *savename( char *s ) {
        char *ap;
        if ((ap = (char *) tcob_malloc( strlen(s) + 1 )) != NULL)
                strcpy(ap,s);
        return( ap );
}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                          subr_lookup                                   |
 |                                                                        |
\*------------------------------------------------------------------------*/

static struct subr_desc *subr_lookup( char *s ) {
        struct subr_desc *sd;
        for ( sd = subrtab[ subr_hash(s) ] ; sd != NULL ; sd = sd->next )
                if ( strcmp( s, sd->name ) == 0 )
                        return sd;
        return NULL;
}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                          subr_install                                  |
 |                                                                        |
\*------------------------------------------------------------------------*/

static struct subr_desc *subr_install( char *name,void *(*subr)() ) {
        int val;
        struct subr_desc *sd=subr_lookup(name);
        if (sd == NULL) {
                sd = tcob_malloc( sizeof(struct subr_desc) );
                if (sd==NULL)
                        return NULL;
                if ( (sd->name = savename( name ) )  == NULL )
                        return NULL;
                val = subr_hash( sd->name );
                sd->next = subrtab[ val ];
                sd->subr = subr;
                subrtab[ val ] = sd;
        }
        return sd;
}
/*------------------------------------------------------------------------*\
 |                                                                        |
 |                          subr_uninstall                                |
 |                                                                        |
\*------------------------------------------------------------------------*/

static void subr_uninstall( char *name) {
        struct subr_desc *sd=NULL;
        struct subr_desc *ant=NULL;
	int hash;
	hash = subr_hash(name);
	
        for ( sd = subrtab[ hash ] ; sd != NULL ; sd = sd->next ){
                if ( strcmp( name, sd->name ) == 0 ){
			if (ant == NULL)
		            subrtab[ hash ] = sd->next;
			else
			    ant->next = sd->next;
			if ( sd->name )free(sd->name);
			if ( sd ) free(sd);
			return;
		}
		ant = sd;
	}
}
/*------------------------------------------------------------------------*\
 |                                                                        |
 |                          module_install                                |
 |                                                                        |
\*------------------------------------------------------------------------*/
static void module_install(void *module){

	struct resolve_handle *rh;

        rh = tcob_malloc(sizeof(struct resolve_handle));
     	rh->handle = module;
        rh->next = resolve_handles;

        resolve_handles = rh;

}
/*------------------------------------------------------------------------*\
 |                                                                        |
 |                          module_uninstall                              |
 |                                                                        |
\*------------------------------------------------------------------------*/
static void module_uninstall(void *module){

	struct resolve_handle *rh;
	struct resolve_handle *ant=NULL;
        for (rh=resolve_handles; rh != NULL; rh=rh->next) {
	     	if (rh->handle == module) {
			if(ant)
				ant->next = rh->next;
			else
				resolve_handles = rh->next;
			if ( rh ) free(rh);
			return ;
		}
		ant = rh;
	}	

}
/*------------------------------------------------------------------------*\
 |                                                                        |
 |                          chkmalloc                                     |
 |                                                                        |
\*------------------------------------------------------------------------*/

/*static void * chkmalloc( size_t size ) {

        void *ptr=malloc(size);
        if (ptr == NULL) {
                fprintf(stderr,"*** Fatal error in malloc(%d)\n",size);
        }
        return ptr;
}*/

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                     tcob_resolve_subr_error                            |
 |                                                                        |
\*------------------------------------------------------------------------*/

void tcob_resolve_subr_error(void) {

        fprintf(stderr,"*** RT Error: dynamic library call \"%s\" not found!\n",
                subrname);
}

/*
 *
 * tcob_resolve_in_previous_libs 
 * 
 * Try to resolve a routine handler on previous opened libs
 *
 *
 */
static void *tcob_resolve_in_previous_libs(char *subrname){

        struct resolve_handle *rh;
        
#ifdef USE_LTDL_LIB
	lt_ptr _dynsubr = NULL;
#else	
        void *(* _dynsubr)()=NULL;
#endif	

        for (rh=resolve_handles; rh != NULL; rh=rh->next) {

#ifdef USE_LTDL_LIB
                _dynsubr = lt_dlsym(rh->handle,subrname);
#else	
                _dynsubr = dlsym(rh->handle,subrname);
#endif	

                if (_dynsubr != NULL)  {
                        subr_install(subrname,_dynsubr);
                        return _dynsubr;
                }
        }
	return NULL;

}
/*
 *
 * tcob_cancel_subr
 *
 * Cancel a dynamically loaded module and uninstall it.
 *
 */
void tcob_cancel_subr(char *name){
	struct resolve_handle *rh;
#ifdef USE_LTDL_LIB
	lt_ptr _dynsubr = NULL;
#else	
        void *(* _dynsubr)()=NULL;
#endif	

        for (rh=resolve_handles; rh != NULL; rh=rh->next) {
#ifdef USE_LTDL_LIB
                _dynsubr = lt_dlsym(rh->handle,subrname);
#else	
                _dynsubr = dlsym(rh->handle,subrname);
#endif	
                if (_dynsubr)  {
#ifdef USE_LTDL_LIB
	        	lt_dlclose(rh->handle);	
#else	
	        	dlclose(rh->handle);	
#endif	
			module_uninstall(rh->handle);
			break;
                }
        }
        subr_uninstall(name);
}
/*
 * tcob_find_in_lib
 *
 * Try to find a library handler on a library name
 *
 */
static void *tcob_find_in_lib(char *rutname,char *libname){

#ifdef USE_LTDL_LIB
        lt_dlhandle module;        
        lt_ptr _dynsubr = NULL;
#else        
        void *(* _dynsubr)()=NULL;
        void *module;
#endif	
        //fprintf(stderr,"findinf '%s' in '%s'\n",rutname,libname);
	// open the library
#ifdef USE_LTDL_LIB
        module = lt_dlopenext(libname);
#else        
        module = dlopen(libname,RTLD_LAZY);
#endif	
	
        if (module != NULL) {
		
#ifdef USE_LTDL_LIB
                _dynsubr = lt_dlsym(module,rutname);
#else        
                _dynsubr = dlsym(module,subrname);
#endif	

                if (_dynsubr != NULL)  {
                        subr_install(rutname,_dynsubr);
                	// save for speeding up next search 
			module_install(module);
			// return the resolved handler
                        return _dynsubr;
                } 

#ifdef USE_LTDL_LIB
       		lt_dlclose(module);
#else        
        	dlclose(module);
#endif	
	} else {
//        	fprintf(stderr,"error '%s' \n",lt_dlerror());
	}
	return NULL;
}

/*
 * tcob_resolve 
 *
 * try to resolve a rutine name, in all the paths specified by TCOB_LD_LIBRARY_PATH
 * 
 */
void *tcob_resolve(char *rutname){

#ifdef USE_LTDL_LIB
	lt_ptr _dynsubr = NULL;
#else	
        void *(* _dynsubr)()=NULL;
#endif	
#ifndef __MINGW32__	
        glob_t g;
	char **libname;
        char *ldlib = NULL ,*libs1 = NULL,*libstmp=NULL;
#endif	
//        struct subr_desc *sd = NULL;
        char *libs = NULL;

        // check if we've not seen this before 
        //if ((sd=subr_lookup(subrname))!=NULL) {
        //        return (sd->subr);
        //}

        // search our currently linked libraries 
	_dynsubr = tcob_find_in_lib(subrname,NULL);
	if (_dynsubr) { 
		return _dynsubr;
	}

        // search in previously opened library handles 
        _dynsubr = tcob_resolve_in_previous_libs(subrname);
	if (_dynsubr) { 
		return _dynsubr;
	}
        // First find a library named as the routine on all the library path
	_dynsubr = tcob_find_in_lib(subrname,subrname);
	if (_dynsubr) { 
		return _dynsubr;
	}
	// find rutname.so 
	libs = tcob_malloc(strlen(subrname) + strlen(HTCOB_DYNLIBS_SUFIX) + 1);
        strcpy(libs,subrname);
        strcat(libs,HTCOB_DYNLIBS_SUFIX);
	_dynsubr = tcob_find_in_lib(subrname,libs);
        free(libs);
	if (_dynsubr) { 
		return _dynsubr;
	}
	// find ./rutname.so 
	libs = tcob_malloc(strlen(subrname)+strlen(HTCOB_DYNLIBS_SUFIX)+3);
        strcpy(libs,".");
        strcat(libs,HTCOB_PATH_DELIM_STR);
        strcat(libs,subrname);
        strcat(libs,HTCOB_DYNLIBS_SUFIX);
	_dynsubr = tcob_find_in_lib(subrname,libs);
        free(libs);
	if (_dynsubr) { 
		return _dynsubr;
	}
	// find ./librutname.so 
	libs = tcob_malloc(strlen(subrname)+strlen(HTCOB_DYNLIBS_SUFIX)+6);
        strcpy(libs,".");
        strcat(libs,HTCOB_PATH_DELIM_STR);
        strcat(libs,"lib");
        strcat(libs,subrname);
        strcat(libs,HTCOB_DYNLIBS_SUFIX);
	_dynsubr = tcob_find_in_lib(subrname,libs);
        free(libs);
	if (_dynsubr) { 
		return _dynsubr;
	}

#ifndef __MINGW32__	
        // Now find the modules on the current dir and on the TC library PATH 
        // setup our libraries names to search (in current directory) 
        g.gl_offs = 0;
        glob(libs_pattern, 0,NULL,&g);
        // add libraries found in HTCOB_LD_PATH 
        libs1 = getenv(HTCOB_LD_PATH);
        if (libs1 != NULL) {
                libstmp = tcob_malloc(strlen(libs1)+1);
                strcpy(libstmp,libs1);
                libs1=libstmp;
        }
        // find all the libraryes on the HTCOB_LD_PATH	
        while ((libs=libs1)!=NULL) {
                if ((libs1 = strchr(libs,HTCOB_PATH_SEP_CHAR)) != NULL) {
                        *libs1++ = 0;
                }
                ldlib = tcob_malloc(strlen(libs)+strlen(libs_pattern)+2);
                strcpy(ldlib,libs);
		
                strcat(ldlib,HTCOB_PATH_DELIM_STR);
                strcat(ldlib,libs_pattern);
		
                glob(ldlib, GLOB_APPEND, NULL,&g);
                free(ldlib);
        }

        if (libstmp != NULL)
                free(libstmp);

        // loop until we find a function with the name given 
        libname = g.gl_pathv;
        if (libname != NULL) {
                for ( ; *libname!=NULL; libname++) { 
		       _dynsubr = tcob_find_in_lib(subrname,*libname);
                       if (_dynsubr != NULL)  {
			       return _dynsubr;
		       }		
                }
        }
        globfree(&g);
#endif	
        return NULL;

}
/*------------------------------------------------------------------------*\
 |                                                                        |
 |                         tcob_resolve_subr                              |
 |                                                                        |
\*------------------------------------------------------------------------*/


void *tcob_resolve_subr(struct fld_desc *f, char *s) {
        
	char *p;

#ifdef USE_LTDL_LIB
        if (!dl_initialized)
        {
                lt_dlinit();
                dl_initialized++;
        }
#endif
        // get our subroutine name to call 
        strncpy(subrname,s,f->len);
        subrname[f->len]=0;
	
        // remove unnecessary spaces 
        p = strchr (subrname, ' ');
        if (p)
          *p = '\0';

        // change - by _ on sub name	
	p=subrname;
	while(*p){
              if (*p=='-') *p='_';
	      p++;
	}

	return tcob_resolve(subrname);
}
/* end of dyncall.c */
