// 
// Copyright (C) 2001, 2000, 1999,  Rildo Pragana, Jim Noeth, 
//               Andrew Cameron, David Essex.
// Copyright (C) 1993, 1991  Rildo Pragana.
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation; either version 2.1,
// or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; see the file COPYING.LIB.  If
// not, write to the Free Software Foundation, Inc., 59 Temple Place,
// Suite 330, Boston, MA 02111-1307 USA


//
//  TinyCOBOL Run Time Library 
//  Dynamic CALL statement functions
//

#define SUBR_HASHLEN            128

#include "htcoblib.h"
#include "rterrors.h"


#ifdef USE_LTDL_LIB
# include <ltdl.h>
#else
# include "ltdl.h"
#endif // HAS_LTDL

#ifndef __MINGW32__
#include <glob.h>
#endif

#ifdef __WINDOWS__
#define HTCOB_DYNLIBS_PATTERN   "*.dll"
#else
#define HTCOB_DYNLIBS_PATTERN   "lib*.so"
#endif

static char subrname[65]; /* allow at most 64 chars subr names */
static int dl_initialized = 0;
static struct resolve_handle {
	void *handle;
	struct resolve_handle *next;
} *resolve_handles=NULL;

/* 
   this mini-symtable will register previously called subroutines 
   to speed up resolving their names to pointers 
*/
static struct subr_desc {
	struct subr_desc *next;
	char *name;
	void *(*subr)();
} *subrtab[ SUBR_HASHLEN ]={NULL};

static char libs_pattern[]=HTCOB_DYNLIBS_PATTERN;
	

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                          subr_hash                                     |
 |                                                                        |
\*------------------------------------------------------------------------*/

static int subr_hash( char *s ) {
	int val=0, i;
	for (i=0; s[i]!='\0'; i++)
		val += s[i];
	return( val % SUBR_HASHLEN );
}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                          savename                                      |
 |                                                                        |
\*------------------------------------------------------------------------*/

static char *savename( char *s ) {
	char *ap;
	if ((ap = (char *) malloc( strlen(s) + 1 )) != NULL)
		strcpy(ap,s);
	return( ap );
}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                          subr_lookup                                   |
 |                                                                        |
\*------------------------------------------------------------------------*/

static struct subr_desc *subr_lookup( char *s ) {
	struct subr_desc *sd;
	for ( sd = subrtab[ subr_hash(s) ] ; sd != NULL ; sd = sd->next )
		if ( strcmp( s, sd->name ) == 0 )
			return sd;
	return NULL;
}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                          subr_install                                  |
 |                                                                        |
\*------------------------------------------------------------------------*/

static struct subr_desc *subr_install( char *name,void *(*subr)() ) {
	int val;
	struct subr_desc *sd=subr_lookup(name);
	if (sd == NULL) {
		sd = malloc( sizeof(struct subr_desc) );
		if (sd==NULL)
			return NULL;
		if ( (sd->name = savename( name ) )  == NULL )
			return NULL;
		val = subr_hash( sd->name );
		sd->next = subrtab[ val ];
		sd->subr = subr;
		subrtab[ val ] = sd;
	}
	return sd;
}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                          tcob_resolve_subr_error                            |
 |                                                                        |
\*------------------------------------------------------------------------*/

void tcob_resolve_subr_error( ) {
	fprintf(stderr,"*** RT Error: dynamic library call \"%s\"not found!\n",
		subrname);
}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                          chkmalloc                                     |
 |                                                                        |
\*------------------------------------------------------------------------*/

static void * chkmalloc( size_t size ) {
	void *ptr=malloc(size);
	if (ptr == NULL) {
		fprintf(stderr,"*** Fatal error in malloc(%d)\n",size);
	}
	return ptr;
}

/*------------------------------------------------------------------------*\
 |                                                                        |
 |                          tcob_resolve_subr                              |
 |                                                                        |
\*------------------------------------------------------------------------*/

void * tcob_resolve_subr( struct fld_desc *f, char *s ) {
	
	//void *(* _dynsubr)()=NULL;
	struct resolve_handle *rh;
	void *handle;
	char **libname;
	struct subr_desc *sd;
	lt_dlhandle module;	
	lt_ptr _dynsubr = NULL;
	char *ldlib,*libs,*libs1,*libstmp=NULL;
#ifndef __MINGW32__
	glob_t g;
#endif

	if (!dl_initialized)
	{
		lt_dlinit();
		libs1 = getenv(HTCOB_LD_PATH);
		if (libs1){
		    lt_dlsetsearchpath(libs1);
		    lt_dladdsearchdir(".");
		} else {
		    lt_dlsetsearchpath(".");	
		}
		dl_initialized++;
	}
#ifndef __MINGW32__
	g.gl_offs = 0;
#endif
	// get our subroutine name to call 
	strncpy(subrname,s,f->len);
	subrname[f->len]=0;
	// remove unnecessary spaces 
	s = strchr (subrname, ' ');
	if (s)
	  *s = '\0';
	// check if we've not seen this before 
	if ((sd=subr_lookup(subrname))!=NULL) {
		return (sd->subr);
	}
	// search our currently linked libraries 
	module = lt_dlopenext(NULL);
	if (module != NULL) {
		_dynsubr = lt_dlsym(module,subrname);
		if (_dynsubr != NULL)  {
			subr_install(subrname,_dynsubr);
			return _dynsubr;
		}
	}
	// search in previously opened library handles 
	for (rh=resolve_handles; rh != NULL; rh=rh->next) {
		_dynsubr = lt_dlsym(rh->handle,subrname);
		if (_dynsubr != NULL)  {
			subr_install(subrname,_dynsubr);
			return _dynsubr;
		}
	}
	// First find a library named as the routine on all the library path
	module = lt_dlopenext(subrname);
	if (module != NULL) {
		_dynsubr = lt_dlsym(module,subrname);
		if (_dynsubr != NULL)  {
			subr_install(subrname,_dynsubr);
			return _dynsubr;
		} else {
			cob_rt_error("dyncall_ltdl: ",301,subrname,lt_dlerror());
		}
	}
#ifndef __MINGW32__
	// Now find the modules on the current dir and on the TC library PATH 
	// setup our libraries names to search (in current directory) 
	glob(libs_pattern, 0,NULL,&g);
	// add libraries found in HTCOB_LD_PATH 
	libs1 = lt_dlgetsearchpath();
	if (libs1 != NULL) {
		libstmp = chkmalloc(strlen(libs1)+1);
		strcpy(libstmp,libs1);
		libs1=libstmp;
	}
	while ((libs=libs1)!=NULL) {
		if ((libs1 = strchr(libs,':')) != NULL) {
			*libs1++ = 0;
		}
		// Add routines on the TC library PATH to the file list
		ldlib = chkmalloc(strlen(libs)+strlen(libs_pattern)+2);
		strcpy(ldlib,libs);
		strcat(ldlib,"/");
		strcat(ldlib,libs_pattern);
		glob(ldlib, GLOB_APPEND, NULL,&g);
		free(ldlib);
	}
	if (libstmp != NULL)
		free(libstmp);
	// loop until we find a function with the name given 
	libname = g.gl_pathv;
	if (libname != NULL) {
		for ( ; *libname!=NULL; libname++) { 
			handle = lt_dlopen(*libname);
			if (handle != NULL) {
				_dynsubr = lt_dlsym(handle,subrname);
				if (_dynsubr != NULL)  {
					// save for speeding up next search 
					rh = chkmalloc(sizeof(struct resolve_handle));
					rh->handle = handle;
					rh->next = resolve_handles;
					resolve_handles = rh;
					break;
				} else { 
					lt_dlclose(handle);
				}
			} else {
			  cob_rt_error("dyncall_ltdl: ",301,subrname,lt_dlerror());
			}
		}
	}
	globfree(&g);
#endif
	subr_install(subrname,_dynsubr);
	return _dynsubr;
}


// end of dyncall.c 
