// 
// Copyright (C) 2002,2001, 2000, 1999,  Rildo Pragana, Jim Noeth, 
//               Andrew Cameron, David Essex.
// Copyright (C) 1993, 1991  Rildo Pragana.
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation; either version 2.1,
// or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this library; see the file COPYING.LIB.  If
// not, write to the Free Software Foundation, Inc., 59 Temple Place,
// Suite 330, Boston, MA 02111-1307 USA


//
//  Cobol Compiler Run Time Library 
//  Curses Screen I/O Functions - Screen frontedn
//

#include "htcoblib.h"
#include "mwindows.h"

#if defined(SunOS)
#include <curses.h>
#else
#  if defined(__CYGWIN__)
#  include <ncurses/ncurses.h>
#  else
#    if defined(__MINGW32__)
#    include <pdcurses.h>
#    else
#    include <curses.h>
#    endif
#  endif
#endif

#include "screenio.h"
#include "rterrors.h"
/* define a macro for CTRL modifier on keys */
#ifndef CTRL
#define CTRL(x) ((x) & 0x1f)
#endif
/* PDcurses 2.4 does not have the following defenetion */
#ifndef KEY_CODE_YES
#define KEY_CODE_YES	0x100	/* A wchar_t contains a key code */
#endif



/* Struct to bind the key pressed and the action to do */
struct KeyBinding {
  int keyCode;		/* Key code returned by getch */
  short action;		/* Action to perform */
  int screenStatus;	/* Value to move to the SCREEN STATUS */
};

/* Table to bind all the keys to actions */
struct KeyBinding keyBindings[] = {
/*   Key Code   , Action on edit , screen status   */   	
  {TCOBKEY_ESCAPE, EDIT_TERMINATE, TCOBKEY_ESCAPE}, /* Escape Key */
  {TCOBKEY_RETURN, EDIT_TERMINATE, 13}, /* For RETURN, we finish the edit mode */
  {KEY_ENTER, EDIT_TERMINATE, 13},     /* ENTER does the same as RETURN */ 

  {KEY_LEFT, EDIT_LEFT, KEY_LEFT},
  {KEY_RIGHT, EDIT_RIGHT, KEY_RIGHT},
  {KEY_HOME, EDIT_FIRST, KEY_HOME},
  {KEY_END, EDIT_LAST, KEY_END},

  {KEY_UP, EDIT_UP_FIELD, KEY_UP},
  {KEY_DOWN, EDIT_DOWN_FIELD, KEY_NEXT},

  {TCOBKEY_TAB, EDIT_NEXT_FIELD, TCOBKEY_TAB},      /* Tab */
  {KEY_BTAB, EDIT_PREV_FIELD, KEY_BTAB},
  {CTRL('n'), EDIT_NEXT_FIELD, 14},
  {CTRL('p'), EDIT_PREV_FIELD, 16},

  {CTRL('f'), EDIT_FIRST_FIELD, 6},
  {CTRL('l'), EDIT_LAST_FIELD, 12},

  {KEY_DC, EDIT_DELETE, KEY_DC},
  {KEY_IC, EDIT_INSERT, KEY_IC},
  {KEY_BACKSPACE, EDIT_BACKSPACE, KEY_BACKSPACE},

  {KEY_PPAGE, EDIT_TERMINATE, KEY_PPAGE},
  {KEY_NPAGE, EDIT_TERMINATE, KEY_NPAGE},
  {KEY_F(1), EDIT_TERMINATE, KEY_F(1)},
  {KEY_F(2), EDIT_TERMINATE, KEY_F(2)},
  {KEY_F(3), EDIT_TERMINATE, KEY_F(3)},
  {KEY_F(4), EDIT_TERMINATE, KEY_F(4)},
  {KEY_F(5), EDIT_TERMINATE, KEY_F(5)},
  {KEY_F(6), EDIT_TERMINATE, KEY_F(6)},
  {KEY_F(7), EDIT_TERMINATE, KEY_F(7)},
  {KEY_F(8), EDIT_TERMINATE, KEY_F(8)},
  {KEY_F(9), EDIT_TERMINATE, KEY_F(9)},
  {KEY_F(10), EDIT_TERMINATE, KEY_F(10)},
  {KEY_F(11), EDIT_TERMINATE, KEY_F(11)},
  {KEY_F(12), EDIT_TERMINATE, KEY_F(12)},
  {0, 0, 0}
};


static short fore_color;      //  Actual fore and back colors	
static short back_color;

static int cursor_x = 0;      // Actual screen position for dinamically 
static int cursor_y = 0;      // calculate +/- line/col                 
//
// Initialize curses 
//
int tcob_scr_init(){
	int i;
    	initscr();
	noecho();
	cbreak();
	keypad(stdscr,TRUE);
	scrollok(stdscr,TRUE);
	nonl();
	//raw();
	noecho();
	if (has_colors())
	    start_color();

	pair_content(0,&fore_color,&back_color);
	attrset(A_NORMAL);

        //
        // This is needed for pdcurses, that has all the pairs to 0-7
        //
        for(i=0;i<COLOR_PAIRS;i++)
           init_pair(i,0,0);

	return TRUE;

}
//
// Close curses
//
void tcob_scr_finish(){
	if(!isendwin())
	     endwin();
}
//
// Set len positions to c
//
static void tcob_scr_setchar(char c, unsigned int len){
	unsigned int i;
	for (i=0; i<len; i++)
		addch(c);
}
//
//  Get actual cursos position on the screen
//
int tcob_scr_getx(){
	return cursor_x;
}
int tcob_scr_gety(){
	return cursor_y;
}
//
// Set cursos position on the screen
//
void tcob_scr_setyx(int y, int x){
	move(y,x);
	getyx(stdscr,cursor_y,cursor_x);
}
void tcob_scr_setx(int x){
	tcob_scr_setyx(tcob_scr_gety(),x);
}
void tcob_scr_sety(int y){
	tcob_scr_setyx(y,tcob_scr_getx());
}
//
// Get the status of keyboard. 
//
//  return TRUE if there is a char in the buffer and FALSE if not
//
int tcob_scr_kbdstatus(){
	int c;
	nodelay(stdscr,TRUE);
	c=getch();
	nodelay(stdscr,FALSE);
	if (c==ERR) {
		return 0;
	}
	ungetch(c);
	return 1;
}
/*-------------------------------------------------------------------------*\
 |                                                                         |
 |                          tcob_scr_action_for_key                        |
 |  Looks in the table which is the action to return for the key passed.   |
 |  If exists the SCREEN STATUS field moves the value in the table to it.  |
 |                                                                         |
\*-------------------------------------------------------------------------*/

int tcob_scr_action_for_key(int key,int *status) {
	struct KeyBinding *binding=NULL;
	struct KeyBinding *pbindings;
    
	
	for(pbindings=keyBindings; pbindings->keyCode!=0; pbindings++) {
		if (pbindings->keyCode == key) {
			binding=pbindings;
			break;
		}
	}
	if (binding==NULL) {
		return EDIT_NOP;
	}
	*status = binding->screenStatus;
	return binding->action;

}
//
// Returns the curses color code from the cobol color number 
//
static int get_curses_color(int cobol_color) {
	// Color numbers from standard 
 	//
	// 0 - black
	// 1 - blue
	// 2 - green
	// 3 - cyan
	// 4 - red
	// 5 - magenta
	// 6 - brown/yellow
	// 7 - white
	//
	// Curses colors
	// #define COLOR_BLACK	0
	// #define COLOR_RED	1
	// #define COLOR_GREEN	2
	// #define COLOR_YELLOW	3
	// #define COLOR_BLUE	4
	// #define COLOR_MAGENTA	5
	// #define COLOR_CYAN	6
	// #define COLOR_WHITE	7
	//
	
	switch(cobol_color % 8)	{
		case TC_BLACK: 	return COLOR_BLACK;
		case TC_BLUE: 	return COLOR_BLUE;
		case TC_GREEN: 	return COLOR_GREEN;
		case TC_CYAN: 	return COLOR_CYAN;
		case TC_RED: 	return COLOR_RED;
		case TC_MAGENTA:return COLOR_MAGENTA;
		case TC_YELLOW: return COLOR_YELLOW;
		case TC_WHITE:	return COLOR_WHITE;
		default:return 0;
	}	

}
//
// sets the color for the next display
//
static void set_colors(){
	int i;
	short fore,back;
	for(i = 0; i < COLOR_PAIRS; ++i) {
		pair_content(i,&fore,&back);
		if ((fore == fore_color) && (back == back_color)) {
			break;
		}
		if ((fore == 0) && (back == 0)) {
			init_pair(i,fore_color,back_color);
			break;
		}
	}
	if (i != COLOR_PAIRS) 
	{
#ifdef CURSES_HAS_COLOR_SET
		color_set(i,NULL);
#else
                attrset(COLOR_PAIR(i));
#endif
                bkgdset(COLOR_PAIR(i));
               
        } else {
                attrset(A_NORMAL);
	}
	
}
/*-------------------------------------------------------------------------*\
 |                                                                         |
 |                            set_attributes                               |
 |                                                                         |
\*-------------------------------------------------------------------------*/
static void set_atributes(struct ScrFld *pFld) {
	if (pFld->iAttributes != -1)
	{
		if (pFld->iAttributes & SCR_HIGHLIGHT)	attron(A_BOLD);
		else					attroff(A_BOLD);
			
		if (pFld->iAttributes & SCR_LOWLIGHT)	attron(A_DIM);
		else					attroff(A_DIM);
			
		if(pFld->iAttributes & SCR_UNDERLINE)	attron(A_UNDERLINE);
		else					attroff(A_UNDERLINE);			
			
		if(pFld->iAttributes & SCR_REVERSE_VIDEO)	attron(A_REVERSE);
		else					attroff(A_REVERSE);			
			
		if(pFld->iAttributes & SCR_BLINK)	attron(A_BLINK);
		else					attroff(A_BLINK);			
			
		if(pFld->iAttributes & SCR_BELL)
			beep();
		if(pFld->iAttributes & SCR_BLANK_SCREEN) {
			clear();
			tcob_scr_setyx(tcob_scr_gety(),tcob_scr_getx());
		}
	}
	if (pFld->iFgColor != -1){
		fore_color=get_curses_color(pFld->iFgColor);
	}
	if (pFld->iBgColor != -1){
		back_color=get_curses_color(pFld->iBgColor);
	}
	set_colors();
}
/*-------------------------------------------------------------------------*\
 |                                                                         |
 |                            _DisplayField                                |
 |                                                                         |
\*-------------------------------------------------------------------------*/
static void _DisplayField(struct ScrFld *pFld) {
	set_atributes(pFld);
	tcob_scr_setyx(pFld->iLine, pFld->iCol);
	
	if(pFld->iAttributes & SCR_BLANK_LINE){ // erase all the current line
		move(tcob_scr_gety(),0);
		clrtoeol();                
		tcob_scr_setyx(tcob_scr_gety(),tcob_scr_getx());
	}
	if(pFld->iAttributes & SCR_ERASE_EOL) // erase to end of line
		clrtoeol();                
	if(pFld->iAttributes & SCR_ERASE_EOS) // erase to end of screen
		clrtobot();                
	if(pFld->iAttributes & SCR_BELL)
		beep();
	if(pFld->iAttributes & SCR_BLANK_SCREEN){
		clear();
		tcob_scr_setyx(tcob_scr_gety(),tcob_scr_getx());
	}

	if(pFld->iAttributes & SCR_SECURE) {
		tcob_scr_setchar('*', pFld->fldScr.len);
	} else {
		if((pFld->iAttributes & SCR_BLANK_WHEN_ZERO) != SCR_BLANK_WHEN_ZERO) {
			addnstr(pFld->caScr, pFld->fldScr.len);
		} else {
			unsigned int i, sign;
			tcob_scr_setyx(pFld->iLine, pFld->iCol);
			sign = tcob_extract_sign(&pFld->fldWk, pFld->caWk);
			for(i = 0; i < pFld->fldWk.len; ++i) {
				if(pFld->caWk[i] != '0')
					break;
			}
			tcob_put_sign(&pFld->fldWk, pFld->caWk, sign);
			if(i == pFld->fldWk.len)
				tcob_scr_setchar(' ', pFld->fldScr.len);
			else
				addnstr(pFld->caScr, pFld->fldScr.len);
		}
	}
	refresh();
}
/*-------------------------------------------------------------------------*\
 |                                                                         |
 |                            _DisplayForAccept                            |
 |                                                                         |
\*-------------------------------------------------------------------------*/
static void _DisplayForAccept(struct ScrFld *pFld) {

	set_atributes(pFld);
	tcob_scr_setyx(pFld->iLine, pFld->iCol);

	if(pFld->iAttributes & SCR_BLANK_LINE)
		clrtoeol();                

	if(pFld->iAttributes & SCR_SECURE)
		tcob_scr_setchar('*', pFld->fldScr.len);
	else if(pFld->iAttributes & SCR_NOECHO)
		tcob_scr_setchar(' ', pFld->fldScr.len);
	else	
		addnstr(pFld->caWk, pFld->fldScr.len);

}
void tcob_scr_display_field(struct ScrFld *pFld){
	_DisplayField(pFld);
}
void tcob_scr_display_for_input(struct ScrFld *pFld){
	_DisplayForAccept(pFld);
}
void tcob_scr_display(struct ScrFld *fields){
	struct ScrFld *pFld;
		
	for(pFld=fields; pFld; pFld=pFld->pNext) {
	//	if(pFld->fldFrom) {				/* if we have a from field */
	//		tcob_move(pFld->fldFrom, pFld->caFrom, &(pFld->fldScr), pFld->caScr);
	//	}
		tcob_scr_display_field(pFld);
		tcob_scr_setx(tcob_scr_getx() + pFld->fldScr.len);
	}
}

/*-------------------------------------------------------------------------*\
 |                                                                         |
 | move_cursor_first                                                       |
 | move_cursor_last                                                        |
 | move_cursor_left                                                        |
 | move_cursor_right                                                       |
 |                                                                         |
 | This are helper functions for cob_accept_screen to help moving inside a |
 | field.                                                                  |
\*-------------------------------------------------------------------------*/
static void move_cursor_first(struct ScrFld *pFld) {
	pFld->iScrPos = 0;
	pFld->iFldPos = 0;
}
static void move_cursor_last(struct ScrFld *pFld) {
	pFld->iScrPos = pFld->fldScr.len - 1;
	pFld->iFldPos = pFld->fldWk.len - 1;
}
static int move_cursor_left(struct ScrFld *pFld) {
	char cWk;
	if(pFld->iFldPos) {
		--pFld->iFldPos;
		--pFld->iScrPos;
		if((pFld->iScrPos) && (pFld->fldScr.type == DTYPE_EDITED)) {
			cWk = pFld->caScr[pFld->iScrPos];
			while(pFld->iScrPos) {
				if((cWk != '.')
				&& (cWk != ','))
					break;
				--pFld->iScrPos;
				cWk = pFld->caScr[pFld->iScrPos];
			}
		}
		return 1;
	}
	return 0;
}
static int move_cursor_right(struct ScrFld *pFld) {
	char cWk;
	if(pFld->iFldPos < (int)(pFld->fldWk.len - 1)) {
		++pFld->iScrPos;
		++pFld->iFldPos;
		if(pFld->fldScr.type == DTYPE_EDITED) {
			cWk = pFld->caScr[pFld->iScrPos];
			while(pFld->iScrPos < (int)(pFld->fldScr.len - 1)) {
				if((cWk != '.')
				&& (cWk != ','))
					break;
				++pFld->iScrPos;
				cWk = pFld->caScr[pFld->iScrPos];
			}
		}
		return 1;
	}
	return 0;
}

//
// Accept one field 
//
//
//
int tcob_scr_accept_field(struct ScrFld *pFld,int *status){
	int j, k, iKey,bAtEnd,action;
	move_cursor_first(pFld);
	set_atributes(pFld);
	while(1) {	
		tcob_scr_setyx(pFld->iLine, pFld->iCol + pFld->iScrPos);
		refresh();
		iKey = getch();
		*status=0;
		if(!(iKey & KEY_CODE_YES) && !(iKey < ' ')) {	// data key pressed 
			bAtEnd = 0;
			if(pFld->fldScr.type != DTYPE_ALPHANUMERIC) { // numeric input
				if(((iKey < '0') || (iKey > '9')) && 
				    (iKey != '-') &&
				    (iKey != '+') &&
				    (iKey != '.') &&
				    (iKey != ',')) {
					beep();
					iKey = 0;
				}
			}
			if ( iKey ) {
				// upper & loower
				if (pFld->iAttributes & SCR_UPPER) 
                                    iKey=toupper(iKey);
				if (pFld->iAttributes & SCR_LOWER) 
                                    iKey=tolower(iKey);
				pFld->caWk[pFld->iFldPos] = (char)iKey;
				_DisplayForAccept(pFld);
				if(pFld->iFldPos < (int)(pFld->fldScr.len - 1)) {
					++pFld->iFldPos;
					++pFld->iScrPos;
				} else
					bAtEnd = 1;
			}
			if((bAtEnd) && (pFld->iAttributes & SCR_AUTO)) {
				return EDIT_DOWN_FIELD;
			}
		} else {  // non data key pressed
			action = tcob_scr_action_for_key(iKey,status);
			switch(action) {		
				case EDIT_LEFT:			/* move cursor to left */
					if (!(move_cursor_left(pFld))) {
						if (pFld->iAttributes & SCR_AUTO) {
							return EDIT_UP_FIELD;
						} else {
							beep();
						}		
					}
					break;
				case EDIT_RIGHT:			/* move cursor to right */
						if (!(move_cursor_right(pFld))) {
							if (pFld->iAttributes & SCR_AUTO) {
								return EDIT_DOWN_FIELD;
							} else	{
								beep();
							}
						}
						break;
				case EDIT_FIRST:
					move_cursor_first(pFld);
					break;
				case EDIT_LAST:	
					move_cursor_last(pFld);
					break;
				case EDIT_BACKSPACE:
					if(pFld->iFldPos) {
						--pFld->iFldPos;
						--pFld->iScrPos;
					} else {
						beep();
					}
				case EDIT_DELETE:			/* delete character under cursor */
						j = pFld->fldWk.len - pFld->iFldPos - 1;
						k = pFld->iFldPos;
						memmove(pFld->caWk+k, pFld->caWk+k+1, j);
						pFld->caWk[k + j] = ' ';
					_DisplayForAccept(pFld);
					break;
				case EDIT_INSERT:		/* insert a character before cursor */
						j = pFld->fldWk.len - pFld->iFldPos - 1;
						k = pFld->iFldPos;
						memmove(pFld->caWk+k+1, pFld->caWk+k, j);
						pFld->caWk[k] = ' ';
					_DisplayForAccept(pFld);
					break;
				default:			// replace character under cursor 
					return action;
			}
		}
	}
	return EDIT_TERMINATE;
}
//
// Accept all the fields until finish of function key pressed
// Return the value for the screen_status
//
int tcob_scr_accept(struct ScrFld *fields) {
	int 			key_status,action;
	struct ScrFld		*pFld;
	struct ScrFld		*pFirst=NULL;
	struct ScrFld		*pLast=NULL;
	struct ScrFld		*pPrev=NULL;
	
	for (pFld=fields; pFld; pFld=pFld->pNext) {
		prepare_input_field(pFld);		
		if(pFld->fldTo)	{		// cursor stop 
			if(!pFirst)
				pFirst = pFld;
			pLast = pFld;
		}
	}

	if(!pFirst)
		return 0;

	for(pFld=pFirst;pFld;) {  //main accept loop
	 	if(pFld->fldTo) {
			prepare_input_field(pFld);
			tcob_move(&pFld->fldWk, pFld->caWk, &pFld->fldScr, pFld->caScr);
		}

		tcob_scr_display_for_input(pFld);		
		
		if (pFld != pPrev) { // if new field, put the cursor on position
					   // and display the last accepted
			if (pPrev)
			   tcob_scr_display_field(pPrev);
			
		}
        
		pPrev = pFld;
		action = tcob_scr_accept_field(pFld,&key_status);
		tcob_move(&pFld->fldWk, pFld->caWk, &pFld->fldScr, pFld->caScr);

		switch(action) {
		  case EDIT_FIRST_FIELD:	/* move to first input field */
			pFld = pFirst;
			break;  
		  case EDIT_LAST_FIELD:		/* move to last input field */
			pFld = pLast;
			break;  
		  case EDIT_DOWN_FIELD:		/* move cursor to next input field */
						/* find next input field in chain */
			pFld = get_next_input_field(pFld);
			break;
		  case EDIT_UP_FIELD:		/* move cursor to prev input field */
		                  		/* find prev input field in chain */
			pFld = get_prev_input_field(pFld);
			break;
		  case EDIT_NEXT_FIELD:		/* move cursor to next input field */
			pFld = get_next_input_field(pFld);
			if(!pFld)		/* if last field rotate */
				pFld = pFirst;
			break;
		  case EDIT_PREV_FIELD:		/* move cursor to prev input field */
			               		/* find prev input field in chain */
			pFld = get_prev_input_field(pFld);
			if(!pFld)		/* if last field rotate */
				pFld = pLast;
			break;
		  case EDIT_TERMINATE:		/* we're done */
			pFld = NULL;
			break;
		} // switch  
	} // main loop
	if(pPrev)
		tcob_scr_display_field(pPrev); // Display last accepted field
	refresh();
	return key_status;
}





/* scr_curses.c */
