###  !/bin/sh 
#
#  Shell script to configure TC for MinGW and create the Inno setup script.
#
#  License: TinyCOBOL is covered by GNU General Public License 
#  Copyright (C) 2001-2003  David Essex
#
#  cmd: tconfig.mingw.sh [ options ... ]
#  options: [ -l: language {en/es/fr/it/pt_BR/ default=en} ] 
#           [ -d: 'release date' format: 'YYYY\/MM\/DD' ] 
#           [ -r: 'release number' ] 
#           [ -v: version ]
#

e0a="invalid number of parms..."
e0b="invalid option(s) "
e0c="Shell script to configure TC for MinGW and create the Inno setup script"  
#e0d="cmd: tconfig.mingw.sh [-l language {en/es/fr/it/pt_BR default=en}] [ -r 'release date' ] [ -v version ]"  
e0d="cmd: tconfig.mingw.sh [ options ... ]\n     options:\n\t[ -h (print this help file) ]\n\t[ -l language {en/es/fr/it/pt_BR default=en} ]\n\t[ -m (set MF compatability option on) ]\n\t[ -d 'release date' format: 'YYYY\/MM\/DD' ]\n\t[ -r 'release number' ]\n\t[ -v version ]"  
#
p_print1 ()
{
 for arg in "$@"
 do
#   echo $arg >&2
   printf "$arg\n" >&2
 done
}
#
TCOB_MAJOR_VERSION=0
TCOB_MINOR_VERSION=61
TCOB_PATCH_LEVEL=0
TCOB_RELEASE_VERSION=1
TCOB_VERSION=$TCOB_MAJOR_VERSION.$TCOB_MINOR_VERSION.$TCOB_PATCH_LEVEL
TCOB_RELEASE_DATE='2003\/08\/27'
tcob_version=$TCOB_VERSION
tcob_rdate=$TCOB_RELEASE_DATE
tcob_release=$TCOB_RELEASE_VERSION
#
TCOBPP_MAJOR_VERSION=0
TCOBPP_MINOR_VERSION=41
TCOBPP_PATCH_LEVEL=1
TCOBPP_VERSION=$TCOBPP_MAJOR_VERSION.$TCOBPP_MINOR_VERSION.$TCOBPP_PATCH_LEVEL
TCOBPP_RELEASE_DATE='2002\/10\/04'
tcobpp_version=$TCOBPP_VERSION
tcobpp_rdate=$TCOBPP_RELEASE_DATE
#
with_lang=''
tcob_lang=en
#
#/* #undef  USE_MF_COMPATABILITY */
tcob_mfopt='0'
tcob_mfopt1='\/\* \#undef USE_MF_COMPATABILITY \*\/'
tcob_mfopt2='\#define USE_MF_COMPATABILITY 1'
#
#
# Get information from configure.in file
#
p_init ()
{
# cat configure.in | grep  '^' | cut -f2 -d'='
# t3=`echo $t2 | sed 's.\/.\\\/.g'`
#
TCOB_MAJOR_VERSION=`cat configure.in | grep  '^TCOB_MAJOR_VERSION' | cut -f2 -d'='`
TCOB_MINOR_VERSION=`cat configure.in | grep  '^TCOB_MINOR_VERSION' | cut -f2 -d'='`
TCOB_PATCH_LEVEL=`cat configure.in | grep  '^TCOB_PATCH_LEVEL' | cut -f2 -d'='`
TCOB_VERSION=$TCOB_MAJOR_VERSION.$TCOB_MINOR_VERSION.$TCOB_PATCH_LEVEL
TCOB_RELEASE_DATE=`cat configure.in | grep  '^TCOB_RELEASE_DATE' | cut -f2 -d'=' | sed 's.\/.\\\/.g'`
tcob_version=$TCOB_VERSION
tcob_rdate=$TCOB_RELEASE_DATE
#
TCOBPP_MAJOR_VERSION=`cat configure.in | grep  '^TCOBPP_MAJOR_VERSION' | cut -f2 -d'='`
TCOBPP_MINOR_VERSION=`cat configure.in | grep  '^TCOBPP_MINOR_VERSION' | cut -f2 -d'='`
TCOBPP_PATCH_LEVEL=`cat configure.in | grep  '^TCOBPP_PATCH_LEVEL' | cut -f2 -d'='`
TCOBPP_VERSION=$TCOBPP_MAJOR_VERSION.$TCOBPP_MINOR_VERSION.$TCOBPP_PATCH_LEVEL
TCOBPP_RELEASE_DATE=`cat configure.in | grep  '^TCOBPP_RELEASE_DATE' | cut -f2 -d'=' | sed 's.\/.\\\/.g'`
tcobpp_version=$TCOBPP_VERSION
tcobpp_rdate=$TCOBPP_RELEASE_DATE
#
# echo "debug: TCOB_RELEASE_DATE=$TCOB_RELEASE_DATE, $tcob_rdate"
# echo "debug: TCOBPP_RELEASE_DATE=$TCOBPP_RELEASE_DATE, $tcobpp_rdate"
#
}
#
# Check and set language option 
#
p_check_lang_otion ()
{
#
 case "x$with_lang" in
     x) tcob_lang='en'   ;;
     xen) tcob_lang='en' ;;
     xes) tcob_lang='es' ;;
     xfr) tcob_lang='fr' ;;
     xit) tcob_lang='it' ;;
     xpt_BR) tcob_lang='pt_BR' ;;
     *)  echo "Language option '$with_lang' not available, defaulting to language=en (English)" 
         tcob_lang='en';;
 esac
}
#
# Create PP version and command line help files
#
p_cobpp () 
{
 cat cobpp/tcppversion.h.in | \
 sed \
     -e "s/@TCOBPP_MAJOR_VERSION@/$TCOBPP_MAJOR_VERSION/g" \
     -e "s/@TCOBPP_MINOR_VERSION@/$TCOBPP_MINOR_VERSION/g" \
     -e "s/@TCOBPP_PATCH_LEVEL@/$TCOBPP_PATCH_LEVEL/g" \
     -e "s/@TCOBPP_VERSION@/$tcobpp_version/g" \
     -e "s/@TCOBPP_RELEASE_DATE@/$tcobpp_rdate/g" \
 > cobpp/tcppversion.h
#
# cat cobpp/tcpphelp.h.in | 
# sed 
#     -e "s/@tcob_lang@/$tcob_lang/g" 
# > cobpp/tcpphelp.h
}
#
# Create main compiler version and command line help files
#
p_cob () 
{
 cat compiler/htversion.h.in | \
 sed \
     -e "s/@TCOB_MAJOR_VERSION@/$TCOB_MAJOR_VERSION/g" \
     -e "s/@TCOB_MINOR_VERSION@/$TCOB_MINOR_VERSION/g" \
     -e "s/@TCOB_PATCH_LEVEL@/$TCOB_PATCH_LEVEL/g" \
     -e "s/@TCOB_VERSION@/$tcob_version/g" \
     -e "s/@TCOB_RELEASE_DATE@/$tcob_rdate/g" \
 > compiler/htversion.h
#
}
#
#
# Create config 
#
p_config ()
{
#
 if [ $tcob_mfopt = "0" ]
 then
   cat htconfig.mingw.h | \
   sed \
     -e "s/@tcob_lang@/$tcob_lang/g" \
     -e "s/@tcobpp_lang@/$tcob_lang/g" \
   > htconfig.h
 else
   cat htconfig.mingw.h | \
   sed \
     -e "s/@tcob_lang@/$tcob_lang/g" \
     -e "s/@tcobpp_lang@/$tcob_lang/g" \
     -e "s|$tcob_mfopt1|$tcob_mfopt2|" \
   > htconfig.h
 fi
#
}
#
# Create make files
#
p_main ()
{
#
 cp -i Makefile.mingw Makefile
 cp -i cobpp/Makefile.mingw cobpp/Makefile
 cp -i compiler/Makefile.mingw compiler/Makefile
 cp -i lib/Makefile.mingw lib/Makefile
 cp -i compiler/htcobolrc.mingw compiler/htcobolrc
#
}
#
#
# Create the Inno setup script from template
#
p_isetup () 
{
 t1=`echo ' ' | tr ' ' '\15'`
#     -e "s/$/$t1/" 
# sed -c 
 cat info/isetup/tcobol.iss.in | \
 sed \
     -e "s/@TCOB_VERSION@/$TCOB_MAJOR_VERSION.$TCOB_MINOR_VERSION/g" \
     -e "s/@TCOB_RELEASE_VERSION@/$tcob_release/g" \
     -e "s/$/$t1/" \
 > tcobol.iss
}
#
p_init
#
while getopts d:hl:mr:v: opt
do
  case "$opt"
  in
    d)
      tcob_rdate=$OPTARG
      ;; 
    h)
      p_print1 "$e0c" "$e0d"
      exit 1
      ;; 
    l)
      with_lang=$OPTARG
      p_check_lang_otion
      ;; 
    m)
      tcob_mfopt='1'
      ;; 
    r)
      tcob_release=$OPTARG
      ;; 
    v)
      tcob_version=$OPTARG
      ;; 
    \?)
#      e1="$e0b$opt ..."
      p_print1 "$e0c" "$e0d"
      exit 1
      ;; 
  esac  
done
#
p_main
#
p_config
#
p_cobpp
#
p_cob
#
p_isetup
#

